# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import mock
from django.core.management import call_command
from tests.native.django.test_tools.test_migrations.test_templatetags import (
    DJANGO_TEMPLATE, TRANSIFEX_TEMPLATE)
from transifex.common.console import Color
from transifex.native.django.management.commands.transifex import Command
from transifex.native.django.management.common import TranslatableFile
from transifex.native.tools.migrations.review import (FileReviewPolicy,
                                                      NoopReviewPolicy,
                                                      StringReviewPolicy)
from transifex.native.tools.migrations.save import (BackupSavePolicy,
                                                    NewFileSavePolicy,
                                                    NoopSavePolicy,
                                                    ReplaceSavePolicy)

HTML_SAMPLE_1 = DJANGO_TEMPLATE
HTML_SAMPLE_2 = """
{% load i18n %}
{% load t from transifex %}

{% t "Hello!" %}

{% t "May" _context="month name" %}

{% t counter='something'|length name=name %}
{counter, plural, one {
There is only one {name} object.
} other {
There are {counter} {name} objects.
}}
{% endt %}

<a href="{{ url }}">Text</a>
"""
HTML_COMPILED_1 = TRANSIFEX_TEMPLATE

PYTHON_SAMPLE = """
from something import gettext
from django.utils.translation import ugettext as _, ugettext_lazy as _lazy
from django.utils.translation import something as smth, ungettext as _pl, ungettext_lazy as _lazypl
from django.utils.translation import pgettext as _ctx
from django.utils.translation import pgettext_lazy as _lazyctx
from django.utils.translation import ugettext as ug, to_locale, get_language_info as lang_info
from django.utils import translation as _trans
import django
import django.utils
import django.something_else
from django import utils as _utils


# A comment
aa = 34
bb = aa + round(33.3)

_('This is "nice"')
_('This is %s') % something
_('This is %s and %s %s %s') % (3, something, "lit'eral", r"lit\'eral")
_('This is %s') % (3 * 15)
_('Hello! %(name)s %(last_name)s %(age)s %(gender)s') % {'name': 'You', 'last_name': user.last_name, 'age': 33, 'gender': gender}
_('This is %s and %s') % ("\\'", r"\\'")
_('This is %(foo)s and %(bar)s') % {'foo': foo, 'bar': 'bar'}
_('This is %(foo)s and %(bar)s') % dict(foo=foo, bar='bar')
_('This is %s') % obj.something('else', 'is', 'happening')
_lazy('Hello! %(name)s %(last_name)s %(age)s %(gender)s') % {'name': 'You', 'last_name': user.last_name, 'age': 33, 'gender': gender}

plural = _pl('One fish', plural='Many fishes', number=number)
plural = _lazypl('One fish', plural='Many fishes', number=number)
django.utils.translation.ugettext('Hello!')

_utils.translation.ugettext(**dict(message='Hello!'))
zero = _trans.ugettext('This is the zero')
first = _(message='This is the first')
plural = _pl('This is', 'These are', 3)
plural = _pl('One fish', plural='Many fishes', number=number)
plural = _pl(number=number, plural='Many "fishes', singular="One 'fish")
plural = _pl(number=number, plural='Many fishes', singular="One 'fish")


withcontext1 = _ctx('Some context', 'This is a message')
withcontext2 = _ctx(**dict(context='Some context', message='This is a message'))
withcontext3 = _lazyctx('Some context', 'This is a message')

do_something(
    django.utils.translation.ngettext('Hello!', 'Hellos!', 2),
    _('Aha! %(name)s %(last_name)s') % {'name': 'You', 'last_name': user.last_name},
    some_var,
    another=_('Yay %s %s') % [something, 33]
)

# Add an import statement after function calls
from django import utils as _utils
from django.utils.translation import ngettext as _ng


class MyClass(object):

    def myfunc(self, text1, text2):
        send_message(_('Foo %(t1)s - %(t2)s') % {'t1': text1, 't2': text2})
"""

PYTHON_SAMPLE_MIGRATED = """
from something import gettext
from transifex.native.django import lazyt, t
from django.utils.translation import something as smth
from django.utils.translation import to_locale, get_language_info as lang_info
from django.utils import translation as _trans
import django
import django.utils
import django.something_else
from django import utils as _utils


# A comment
aa = 34
bb = aa + round(33.3)

t('This is "nice"')
t('This is {variable_1}', variable_1=something)
t('This is {variable_1} and {variable_2} {variable_3} {variable_4}', variable_1=3, variable_2=something, variable_3="lit'eral", variable_4="lit'eral")
t('This is {variable_1}', variable_1=(3 * 15))
t('Hello! {name} {last_name} {age} {gender}', name='You', last_name=user.last_name, age=33, gender=gender)
t('This is {variable_1} and {variable_2}', variable_1="'", variable_2='\\'')
t('This is {foo} and {bar}', foo=foo, bar='bar')
t('This is {foo} and {bar}', __txnative_fixme="dict(foo=foo, bar='bar')")
t('This is {variable_1}', variable_1=obj.something('else', 'is', 'happening'))
lazyt('Hello! {name} {last_name} {age} {gender}', name='You', last_name=user.last_name, age=33, gender=gender)

plural = t('{cnt, one {One fish} other {Many fishes}}', cnt=number)
plural = lazyt('{cnt, one {One fish} other {Many fishes}}', cnt=number)
t('Hello!')

t('Hello!')
zero = t('This is the zero')
first = t('This is the first')
plural = t('{cnt, one {This is} other {These are}}', cnt=3)
plural = t('{cnt, one {One fish} other {Many fishes}}', cnt=number)
plural = t("{cnt, one {One 'fish} other {Many \\"fishes}}", cnt=number)
plural = t("{cnt, one {One 'fish} other {Many fishes}}", cnt=number)


withcontext1 = t('This is a message', _context='Some context')
withcontext2 = t('This is a message', _context='Some context')
withcontext3 = lazyt('This is a message', _context='Some context')

do_something(
    t('{cnt, one {Hello!} other {Hellos!}}', cnt=2),
    t('Aha! {name} {last_name}', name='You', last_name=user.last_name),
    some_var,
    another=t('Yay {variable_1} {variable_2}', variable_1=something, variable_2=33)
)

# Add an import statement after function calls
from django import utils as _utils


class MyClass(object):

    def myfunc(self, text1, text2):
        send_message(t('Foo {t1} - {t2}', t1=text1, t2=text2))
"""


PATH_FIND_FILES = ('transifex.native.django.management.utils.base.'
                   'CommandMixin._find_files')
PATH_READ_FILE = ('transifex.native.django.management.utils.base.CommandMixin.'
                  '_read_file')
PATH_PROMPT_FILE = 'transifex.native.tools.migrations.review.ReviewPolicy' \
                   '.prompt_for_file'
PATH_PROMPT_STRING = 'transifex.native.tools.migrations.review' \
                     '.ReviewPolicy.prompt_for_string'
PATH_SAVE_FILE = 'transifex.native.tools.migrations.save.SavePolicy' \
                 '._safe_save'
PATH_PROMPT_START1 = 'transifex.native.tools.migrations.execution' \
                     '.MigrationExecutor._prompt_to_start'


@mock.patch(PATH_PROMPT_START1)
@mock.patch(PATH_READ_FILE)
@mock.patch('os.getcwd')
def test_dry_run_save_none_review0(mock_cur_dir, mock_read,
                                   mock_prompt_to_start1):
    mock_cur_dir.return_value = 'dir1/dir2'
    mock_read.side_effect = [
        HTML_SAMPLE_1,  # 1.html
    ]
    command = Command()
    call_command(command, 'migrate', save_policy='none',
                 review_policy='none', files=['1.html'])
    assert isinstance(command.subcommands['migrate'].executor.save_policy,
                      NoopSavePolicy)
    assert isinstance(command.subcommands['migrate'].executor.review_policy,
                      NoopReviewPolicy)


@mock.patch(PATH_PROMPT_START1)
@mock.patch(PATH_READ_FILE)
@mock.patch(PATH_FIND_FILES)
def test_dry_run_save_none_review(mock_find_files, mock_read,
                                  mock_prompt_to_start1):
    mock_find_files.return_value = [
        TranslatableFile('dir1/dir2', '1.html', 'locdir1'),
        TranslatableFile('dir4/dir5', '1.txt', 'locdir1'),
    ]
    mock_read.side_effect = [
        HTML_SAMPLE_1,  # 1.html
        HTML_SAMPLE_2,  # 1.txt,
    ]
    command = Command()
    call_command(command, 'migrate', save_policy='none',
                 review_policy='none')
    assert isinstance(command.subcommands['migrate'].executor.save_policy,
                      NoopSavePolicy)
    assert isinstance(command.subcommands['migrate'].executor.review_policy,
                      NoopReviewPolicy)


@mock.patch(PATH_PROMPT_START1)
@mock.patch(PATH_SAVE_FILE, return_value=(True, None))
@mock.patch(PATH_PROMPT_STRING)
@mock.patch(PATH_PROMPT_FILE)
@mock.patch(PATH_READ_FILE)
@mock.patch(PATH_FIND_FILES)
def test_new_file_save_file_review(mock_find_files, mock_read,
                                   mock_prompt_file, mock_prompt_string,
                                   mock_save_file,
                                   mock_prompt_to_start1):
    mock_find_files.return_value = [
        TranslatableFile('dir1/dir2', '1.html', 'locdir1'),
        TranslatableFile('dir4/dir5', '1.txt', 'locdir1'),
        TranslatableFile('dir4/dir5', '1.py', 'locdir1'),
    ]
    mock_read.side_effect = [
        HTML_SAMPLE_1,  # 1.html
        HTML_SAMPLE_2,  # 1.txt
        PYTHON_SAMPLE,  # 1.py
    ]
    command = Command()
    call_command(command, 'migrate', save_policy='new', review_policy='file')
    assert isinstance(command.subcommands['migrate'].executor.save_policy,
                      NewFileSavePolicy)
    assert isinstance(command.subcommands['migrate'].executor.review_policy,
                      FileReviewPolicy)

    # The first FileMigration instance that reached the review object
    # should compile to the proper Native syntax (1.html)
    file_migration1 = mock_prompt_file.call_args_list[0][0][0]
    assert file_migration1.compile() == HTML_COMPILED_1

    # The same for 1.py
    file_migration2 = mock_prompt_file.call_args_list[1][0][0]
    assert file_migration2.compile() == PYTHON_SAMPLE_MIGRATED

    # No string review should have taken place
    assert mock_prompt_string.call_count == 0

    # The path and content that reached the save object
    # should have the correct values
    assert mock_save_file.call_args_list[0][0][0] == 'dir1/dir2/1__native.html'
    migration_compile = mock_save_file.call_args_list[0][0][1]
    assert migration_compile() == HTML_COMPILED_1

    assert mock_save_file.call_args_list[1][0][0] == 'dir4/dir5/1__native.py'
    migration_compile = mock_save_file.call_args_list[1][0][1]
    assert migration_compile() == PYTHON_SAMPLE_MIGRATED


@mock.patch(PATH_PROMPT_START1)
@mock.patch(PATH_SAVE_FILE, return_value=(True, None))
@mock.patch(PATH_PROMPT_STRING)
@mock.patch(PATH_PROMPT_FILE)
@mock.patch(PATH_READ_FILE)
@mock.patch(PATH_FIND_FILES)
def test_backup_save_string_review(mock_find_files, mock_read,
                                   mock_prompt_file, mock_prompt_string,
                                   mock_save_file,
                                   mock_prompt_to_start1):
    mock_find_files.return_value = [
        TranslatableFile('dir1/dir2', '1.html', 'locdir1'),
        TranslatableFile('dir4/dir5', '1.txt', 'locdir1'),
    ]
    mock_read.side_effect = [
        HTML_SAMPLE_1,  # 1.html
        HTML_SAMPLE_2,  # 1.txt
    ]
    command = Command()
    call_command(command, 'migrate', save_policy='backup',
                 review_policy='string')
    assert isinstance(command.subcommands['migrate'].executor.save_policy,
                      BackupSavePolicy)
    assert isinstance(command.subcommands['migrate'].executor.review_policy,
                      StringReviewPolicy)

    assert mock_prompt_file.call_count == 0
    assert mock_prompt_string.call_count == 13  # 11 migrated strings

    # The path and content that reached the save object
    # should have the correct values, one for the backup and one for
    # the new content
    assert mock_save_file.call_args_list[0][0][0] == 'dir1/dir2/1.html.bak'
    original_content_getter = mock_save_file.call_args_list[0][0][1]
    assert original_content_getter() == HTML_SAMPLE_1

    assert mock_save_file.call_args_list[1][0][0] == 'dir1/dir2/1.html'
    migration_compile = mock_save_file.call_args_list[1][0][1]
    assert migration_compile() == HTML_COMPILED_1


@mock.patch(PATH_PROMPT_START1)
@mock.patch(PATH_SAVE_FILE, return_value=(True, None))
@mock.patch(PATH_PROMPT_STRING)
@mock.patch(PATH_PROMPT_FILE)
@mock.patch(PATH_READ_FILE)
@mock.patch(PATH_FIND_FILES)
def test_replace_save_string_review(mock_find_files, mock_read,
                                    mock_prompt_file, mock_prompt_string,
                                    mock_save_file,
                                    mock_prompt_to_start1):
    mock_find_files.return_value = [
        TranslatableFile('dir1/dir2', '1.html', 'locdir1'),
        TranslatableFile('dir4/dir5', '1.txt', 'locdir1'),
    ]
    mock_read.side_effect = [
        HTML_SAMPLE_1,  # 1.html
        HTML_SAMPLE_2,  # 1.txt
    ]
    command = Command()
    call_command(command, 'migrate', save_policy='replace',
                 review_policy='string')
    assert isinstance(command.subcommands['migrate'].executor.save_policy,
                      ReplaceSavePolicy)
    assert isinstance(command.subcommands['migrate'].executor.review_policy,
                      StringReviewPolicy)

    assert mock_prompt_file.call_count == 0
    assert mock_prompt_string.call_count == 13

    # The path and content that reached the save object
    # should have the correct values
    assert mock_save_file.call_args[0][0] == 'dir1/dir2/1.html'
    migration_compile = mock_save_file.call_args[0][1]
    assert migration_compile() == HTML_COMPILED_1


@mock.patch('transifex.common.console.Color.echo')
def test_text_migration_template_code(mock_echo):
    """Test the mode that migrates directly given text instead of files
    (Django HTML templates)."""
    command = Command()
    call_command(command, 'migrate', text=DJANGO_TEMPLATE)
    expected = Color.format(
        '\n[high]Transifex Native syntax:[end]\n[green]{}[end]'.format(
            TRANSIFEX_TEMPLATE
        )
    )
    actual = Color.format(mock_echo.call_args_list[1][0][0])
    assert expected == actual

    # Make sure it's idempotent
    mock_echo.reset_mock()
    call_command(command, 'migrate', text=TRANSIFEX_TEMPLATE)
    actual = Color.format(mock_echo.call_args_list[1][0][0])
    assert expected == actual


@mock.patch('transifex.common.console.Color.echo')
def test_text_migration_python_code(mock_echo):
    """Test the mode that migrates directly given text instead of files
    (Python/gettext code)."""
    command = Command()
    call_command(command, 'migrate', text=PYTHON_SAMPLE)
    expected = Color.format(
        '\n[high]Transifex Native syntax:[end]\n[green]{}[end]'.format(
            PYTHON_SAMPLE_MIGRATED
        )
    )
    native = Color.format(mock_echo.call_args_list[1][0][0])
    assert expected == native

    # Make sure it's idempotent
    mock_echo.reset_mock()
    call_command(command, 'migrate', text=PYTHON_SAMPLE_MIGRATED)
    actual = Color.format(mock_echo.call_args_list[1][0][0])
    assert expected == actual



from __future__ import unicode_literals


class Confidence(object):
    """The level of certainty for a particular migration.

    If a migration of a string from a certain framework to Transifex Native
    is very complex, or an edge-case, it can be marked as low-confidence.
    """

    LOW = 1
    HIGH = 2

    @classmethod
    def to_string(cls, confidence):
        """Return a string representation of a confidence int type.

        :param int confidence: the int ID
        :rtype: str
        """
        return 'LOW' if confidence == Confidence.LOW else 'HIGH'


class StringMigration(object):
    """Migration info for a framework's i18n syntax to Transifex Native syntax,
    for a single unit (e.g. a Django template tag or a Python function call).

    Describes the changes that need to be made in order to change
    a template file that uses the framework's i18n syntax to a file that uses
    the Transifex Native syntax.
    """

    def __init__(self, original, new, confidence=Confidence.HIGH):
        """Constructor.

        :param unicode original: the original string in 3rd-party syntax
        :param unicode new: the new string in Transifex Native syntax
        :param int confidence: the level of confidence of this migration
        """
        self.original = ''
        self.new = ''
        self.confidence = confidence
        self.modified = False
        self.update(original, new)

    def update(self, extra_original, extra_new, confidence=None, append=True):
        """Update the string migration, adding a new part in the original
        string and a new part in the new string.

        Usage:
        >>> migration = StringMigration('a', 'b')
        >>> # 'a' -> 'b'
        >>> migration.update('c', 'C')
        >>> # 'ac' -> 'bC'
        >>> migration.update('>', '>>', append=False)
        >>> # '>ac' -> '>>bC'
        >>> migration.update('', '<<', append=True)
        >>> # '>ac' -> '>>bC<<'

        :param unicode extra_original: the additional string to add
            to the original string
        :param unicode extra_new: the additional string to add
            to the new string
        :param int confidence: the new level of confidence of the migration;
            if `None` is provided, the previous confidence is preserved
        :param bool append: if True, the changes will be appended to the end
            of the strings, otherwise they will be prepended in the beginning
        """
        if append:
            self.original = self.original + extra_original
            self.new = self.new + extra_new
        else:
            self.original = extra_original + self.original
            self.new = extra_new + self.new

        self.confidence = confidence or self.confidence or Confidence.HIGH
        self.modified = self.new != self.original

    def revert(self):
        """Revert the migration, making the new string identical to the
        original one."""
        self.new = self.original
        self.confidence = Confidence.HIGH
        self.modified = False

    def __repr__(self):
        return (
            '<StringMigration original="{original}" new="{new}" '
            'modified={modified} confidence={confidence}>'
        ).format(
            original=self.original[:20].replace('\n', '\\n') + '...',
            new=self.new[:20].replace('\n', '\\n') + '...',
            modified=self.modified,
            confidence=Confidence.to_string(self.confidence),
        )


class FileMigration(object):
    """Migration info for a framework's i18n syntax to Transifex Native syntax,
    for a whole file.

    Describes the changes that need to be made in order to change
    a template file that uses the framework's i18n syntax to a file that uses
    the Transifex Native syntax.
    """

    def __init__(self, filename, original_content):
        """Constructor.

        :param unicode filename: the path of the original file
        :param unicode original_content: the actual content of the original file
        """
        self.filename = filename
        self.original_content = original_content
        self.strings = []

    def add_string(self, string_migration):
        """Add a string migration.

        Each string migration represents a distinctive item that was found
        in the file. It may have changes or it may not.

        :param StringMigration string_migration: the migration object to add
        """
        self.strings.append(string_migration)

    def compile(self):
        """Render all changes in a string, resulting to the final migrated file.

        :return: a string that includes the final state of the migrated file
        :rtype: unicode
        """
        return ''.join(
            [(x.new if x.new is not None else '') for x in self.strings]
        )

    def revert(self):
        """Revert all string test_migrations in the file."""
        for string_migration in self.strings:
            string_migration.revert()

    @property
    def modified_strings(self):
        """A list of all StringMigration objects with content that was really
        migrated, for a particular file.

        :rtype: List[StringMigration]
        """
        return [string for string in self.strings if string.modified]

    @property
    def low_confidence_strings(self):
        """A list of all StringMigration objects with migrated content
        that has low confidence, for a particular file.

        :rtype: List[StringMigration]
        """
        return [
            x for x in self.modified_strings
            if x.confidence == Confidence.LOW
        ]



import os
import sys

from transifex.common.console import Color, pluralized, prompt
from transifex.native.tools.migrations.mark import (
    MarkLowConfidenceFilesPolicy, MarkLowConfidenceStringsPolicy,
    NoopMarkPolicy, create_mark_policy)
from transifex.native.tools.migrations.models import Confidence
from transifex.native.tools.migrations.review import (
    REVIEW_ACCEPT_ALL, REVIEW_EXIT, REVIEW_REJECT_ALL, FileReviewPolicy,
    LowConfidenceFileReviewPolicy, LowConfidenceStringReviewPolicy,
    NoopReviewPolicy, StringReviewPolicy, create_review_policy)
from transifex.native.tools.migrations.save import (BackupSavePolicy,
                                                    NewFileSavePolicy,
                                                    NoopSavePolicy,
                                                    ReplaceSavePolicy,
                                                    create_save_policy)

SAVE_POLICY_OPTIONS = {
    NoopSavePolicy.name: 'no changes will be saved\n',
    NewFileSavePolicy.name: 'migrated content will be saved in a new file, '
                            'named <filename>__native.<extension>\n',
    BackupSavePolicy.name: 'migrated content will be saved directly in the '
                           'original file path, and a backup will also be '
                           'saved in <filename>.<extension>.bak\n',
    ReplaceSavePolicy.name: 'migrated content will be saved in the original '
                            'file',
}

REVIEW_POLICY_OPTIONS = {
    NoopReviewPolicy.name: 'everything will be done automatically without '
                           'having a chance to review anything\n',
    FileReviewPolicy.name: 'you get a chance to review each migrated file '
                           'before it is saved\n',
    StringReviewPolicy.name: 'you get a chance to review each string of each '
                             'file before the file is saved\n',
    LowConfidenceFileReviewPolicy.name: 'you get a chance to review each '
                                        'migrated file that includes at least '
                                        'one string that has a low migration '
                                        'confidence\n',
    LowConfidenceStringReviewPolicy.name: 'you get a chance to review each '
                                          'string that has a low migration '
                                          'confidence\n',
}

MARK_POLICY_OPTIONS = {
    NoopMarkPolicy.name: 'nothing will automatically be marked for '
                         'proofreading\n',
    MarkLowConfidenceFilesPolicy.name: 'any migrated file that includes '
                                       'at least one string that has low '
                                       'migration confidence will be marked '
                                       'for proofreading\n',
    MarkLowConfidenceStringsPolicy.name: 'any migrated string that has low '
                                         'confidence will be marked for '
                                         'proofreading\n',
}


class MigrationExecutor(object):
    """Responsible for orchestrating a migration of multiple files
    to Transifex Native syntax.

    It guides the user throughout the process, providing important
    feedback in the console. It supports various "review", "save" and "mark"
    policies, that determine what the user is asked to accept or reject
    before being saved, and the file location the changes are saved in.

    It is agnostic to any specific Python framework. The part that
    does the actual transformation from a specific framework (e.g. Django)
    to Transifex Native is handled by the `file_migrator_func` callable
    that is provided externally through dependency injection.
    """

    def __init__(self, options, file_migrator_func):
        """Constructor.

        :param dict options: the following configuration options of
            the migration:
             - 'files': a list of files to migrate, relative to
                current location
             - 'review_policy' (see REVIEW_POLICY_OPTIONS)
             - 'save_policy' (SAVE_POLICY_OPTIONS)
        :param callable file_migrator_func: a callable that is responsible for
            getting an object (e.g. a TranslatableFile) and returning
            a FileMigration object
        """
        self.options = options
        self.file_migrator_func = file_migrator_func
        self.save_policy = create_save_policy(options['save_policy'])
        self.review_policy = create_review_policy(options['review_policy'])
        self.mark_policy = create_mark_policy(options['mark_policy'])
        self.stats = {
            'processed_files': 0,
            'migrations': [],
            'saved': [],
            'files_marked': 0,
            'strings_marked': 0,
            'errors': [],
        }

    def _comment_format(self, path):
        """Return the comment format suitable for the given file,
        based on its extension.

        :param unicode path: the path to a migration file
        :return: a new string, to use with .format()
        :rtype: unicode
        """
        _, extension = os.path.splitext(path)
        return '# {}\n' if extension == '.py' else '<!-- {} -->'

    def migrate_files(self, files):
        """Search all related files, detect Django i18n translate hooks and
        migrate them to Transifex syntax.

        :param list files: a list of TranslatableFile objects
        """
        files_total = len(files)

        # Ask the user for permission to continue
        self._prompt_to_start(len(files))

        accept_remaining_files = False
        exit_migration = False

        # Loop through each file, migrate, ask for user review if applicable,
        # save to disk if applicable
        for file_cnt, translatable_file in enumerate(files):
            if exit_migration:
                break

            comment_format = self._comment_format(translatable_file.file)
            self.review_policy.set_comment_format(comment_format)
            self.mark_policy.set_comment_format(comment_format)

            Color.echo(
                '\n---- '
                '[[high]{cnt}[end]/[high]{total}[end]] '
                'Migrating [file]{path}[end]...'.format(
                    path=translatable_file.path,
                    cnt=file_cnt + 1,
                    total=files_total,
                )
            )
            self.stats['processed_files'] += 1
            file_migration = self.file_migrator_func(translatable_file)
            if not file_migration:
                continue

            modified_strings = file_migration.modified_strings
            total_modified = len(modified_strings)
            total_low_confidence = len(
                [x for x in modified_strings if x.confidence == Confidence.LOW]
            )
            msg = pluralized(
                '[warn]1[end] [prompt]string was modified[end]',
                '[warn]{cnt}[end] [prompt]strings were modified[end]',
                total_modified,
            )
            Color.echo(
                '{msg}{confidence}'.format(
                    msg=msg,
                    confidence=(
                        ' ([warn]{low}[end] with low confidence)'.format(
                            low=total_low_confidence,
                        )
                        if total_low_confidence else ''
                    )
                )
            )
            if not total_modified:
                continue

            # If the review policy says so, prompt the user for each string
            # If this returns True, it doesn't necessarily mean that the user
            # will be prompted, as the actual review policy may use
            # additional filters, e.g. only prompt for strings with low
            # confidence
            if self.review_policy.should_review_strings():
                reject_remaining_strings = False
                for string_index in range(total_modified):
                    string_migration = modified_strings[string_index]

                    # The rest of the string test_migrations should be reverted
                    # based on the user's choice
                    if reject_remaining_strings:
                        string_migration.revert()

                    # Optionally prompt the user to review the migration
                    else:
                        # Give the user the option to review
                        # May modify `string_migration` in place
                        result = self.review_policy.review_string(
                            string_migration, string_index, total_modified
                        )
                        # The user has chosen to accept the changes
                        # in all remaining strings. Break so that
                        # there will be no more prompts for the rest
                        # of the strings
                        if result == REVIEW_ACCEPT_ALL:
                            break

                        # The user has chosen to reject the changes in all
                        # remaining strings. Set the flag to True, so that
                        # it will revert all changes for the rest strings
                        # in the loop
                        elif result == REVIEW_REJECT_ALL:
                            reject_remaining_strings = True

                        # The user has chosen to exit the migration completely
                        # Break to exit the outer (file) loop
                        elif result == REVIEW_EXIT:
                            exit_migration = True
                            break

            # If the mark policy says so, give it a chance to mark
            # each string for proofread
            if self.mark_policy.should_mark_strings():
                for string_migration in modified_strings:
                    marked = self.mark_policy.mark_string(string_migration)
                    if marked:
                        self.stats['strings_marked'] += 1

            # If the review policy says so, prompt the user for each file
            if accept_remaining_files is False and exit_migration is False:
                result = self.review_policy.review_file(file_migration)

                # The user has chosen to reject all remaining files
                # Break, so that we exit the outer (file) loop
                if result == REVIEW_REJECT_ALL:
                    break

                # The user has chosen to accept all remaining files
                # Set the flag, so that the file review policy won't be used
                # for the remaining file test_migrations
                elif result == REVIEW_ACCEPT_ALL:
                    accept_remaining_files = True

                # The user has chosen to exit the migration completely
                elif result == REVIEW_EXIT:
                    exit_migration = True

            # Skip to the results
            # Break to exit the outer (file) loop
            if exit_migration is True:
                break

            # Give a chance to the mark policy to mark the file for proofread
            marked = self.mark_policy.mark_file(file_migration)
            if marked:
                self.stats['files_marked'] += 1

            # If the save policy says so, save the changes
            if file_migration.modified_strings:
                saved, error_type = self.save_policy.save_file(file_migration)
            else:
                saved, error_type = False, None

            # Update stats
            self.stats['migrations'].append(
                (translatable_file.path, file_migration)
            )
            if saved:
                self.stats['saved'].append(file_migration)
            elif error_type is not None:
                self.stats['errors'].append(file_migration)

        self._show_results(files, self.stats)

    def show_intro(self):
        """Show an introductory message to help the user understand what
        is going on.
        """
        Color.echo(
            '[high]'
            '\n############################################################'
            '########\n'
            'Running migration from Django i18n syntax to Transifex Native '
            'syntax\n'
            '[end]'
            '\nThis migration is idempotent, so its output should not '
            'change if run'
            '\nmultiple times with the same configuration.'
        )
        Color.echo('\n[high]Configuration:[end]')
        if self.options['path']:
            Color.echo(
                '[opt]Path:[end] [file]{}[end]'.format(self.options['path'])
            )
        if self.options['files']:
            Color.echo('[opt]Files:[end]')
            Color.echo(
                '\n'.join([
                    ' - [file]{}[end]'.format(x)
                    for x in self.options['files']
                ])
            )
        Color.echo(
            '[opt]Review policy:[end] [high]{}[end] -> {}'.format(
                self.options['review_policy'],
                REVIEW_POLICY_OPTIONS[self.options['review_policy']],
            ).strip()
        )
        Color.echo(
            '[opt]Save policy:[end] [high]{}[end] -> {}'.format(
                self.options['save_policy'],
                SAVE_POLICY_OPTIONS[self.options['save_policy']],
            ).strip()
        )
        Color.echo(
            '[opt]Mark policy:[end] [high]{}[end] -> {}'.format(
                self.options['mark_policy'],
                MARK_POLICY_OPTIONS[self.options['mark_policy']],
            ).strip()
        )

    def _prompt_to_start(self, total_files):
        """Prompt the user before starting the migration.

        If the user chooses to not go through with it, sys.exit() is called.

        :param int total_files: the total number of files to migrate
        """
        msg = pluralized(
            'Found [warn]{cnt}[end] file to check for translatable strings.',
            'Found [warn]{cnt}[end] files to check for translatable strings.',
            total_files,
        )
        Color.echo('\n{}'.format(msg))

        if not total_files:
            Color.echo('\n[high]Migration ended.[end]')
            sys.exit(1)

        if (
            self.options['save_policy'] != NoopSavePolicy.name
            and self.options['review_policy'] == NoopReviewPolicy.name
        ):
            Color.echo(
                '\n[warn]WARNING! The selected configuration will save '
                'all files automatically, without allowing you to do any '
                'reviewing first.[end]'
            )

        while True:
            reply = prompt(
                Color.format(
                    '[opt](Y)[end] Yes [opt](N)[end] No'
                ),
                description='Are you sure you want to continue?',
                default='N',
            )
            reply = reply.upper()
            if reply == 'Y':
                return
            elif reply == 'N':
                Color.echo('\n[high]Migration aborted.[end]')
                sys.exit(1)

    def _show_results(self, files, stats):
        """Show a detailed report of how the migration went.

        :param list files: a list of TranslatableFile objects
        :param dict stats: a dictionary with all statistics of the execution
        """
        Color.echo('\n\n[high]Migration completed![end]')
        Color.echo('--------------------')
        Color.echo('[high]Files found:[end] [warn]{}[end]'.format(
            len(files))
        )
        Color.echo('[high]Files processed:[end] [warn]{}[end]'.format(
            stats['processed_files'])
        )

        files_modified = 0
        strings_modified = 0
        for _, file_migration in stats['migrations']:
            new_string_count = len(file_migration.modified_strings)
            strings_modified += new_string_count
            if new_string_count:
                files_modified += 1

        # Files & string migrations
        Color.echo(
            '[high]File migrations created:[end] [warn]{}[end]'.format(
                files_modified
            )
        )
        Color.echo(
            '[high]String migrations inside these files: [warn]{}[end]'.format(
                strings_modified
            )
        )

        # Files saved
        Color.echo('[high]Files saved:[end] [warn]{}[end]'.format(
            len(stats['saved'])
        ))
        saved_str = '\n'.join([
            ' - [file]{}[end]'.format(x.filename)
            for x in stats['saved']
        ])
        if saved_str:
            Color.echo(saved_str)

        # Files & strings marked
        Color.echo(
            '[high]Files marked for proofreading:[end] [warn]{}[end]'.format(
                stats['files_marked']
            )
        )
        Color.echo(
            '[high]Strings marked for proofreading:[end] [warn]{}[end]'.format(
                stats['strings_marked']
            )
        )

        # Errors found
        Color.echo('[high]Errors found:[end] [warn]{}[end]'.format(
            len(stats['errors'])
        ))
        errors_str = '\n'.join([
            ' - [warn]{}[end]'.format(x.filename)
            for x in stats['errors']
        ])
        if errors_str:
            Color.echo(errors_str)

        Color.echo('')


def migrate_text(text, migrator_func):
    """Convert the given text from the original framework to Native syntax.

    Supports both HTML/template syntax and Python/gettext syntax.
    Prints out the result in the console.

    :param unicode text: the text to migrate to Native syntax
    :param callable migrator_func: a Callable[unicode] -> FileMigration object
        that converts syntax to Transifex Native; provided externally
        so that it can support any Python framework (e.g. Django)
    """
    Color.echo(
        '[high]Original syntax:[end]\n[red]{}[end]'.format(text)
    )
    file_migration = migrator_func(text)
    Color.echo(
        '\n[high]Transifex Native syntax:[end]\n[green]{}[end]'.format(
            file_migration.compile()
        )
    )



from __future__ import absolute_import, unicode_literals

import json
import os

import transifex.native.consts as consts
from django.conf import settings
from django.core.management.utils import handle_extensions
from django.utils.encoding import force_text
from transifex.common.console import Color
from transifex.native import tx
from transifex.native.django.management.common import SourceStringCollection
from transifex.native.django.management.utils.base import CommandMixin
from transifex.native.django.utils.templates import \
    extract_transifex_template_strings
from transifex.native.parsing import Extractor


class Push(CommandMixin):
    def add_arguments(self, subparsers):
        parser = subparsers.add_parser(
            'push',
            help=("Detect translatable strings in Django templates and Python "
                  "files and push them as source strings to Transifex"),
        )
        parser.add_argument(
            '--extension', '-e', dest='extensions', action='append',
            help=('The file extension(s) to examine (default: "html,txt,py", '
                  'or "js" if the domain is "djangojs"). Separate multiple '
                  'extensions with commas, or use -e multiple times.'),
        )
        parser.add_argument(
            '--purge', '-p', action='store_true', dest='purge', default=False,
            help=('Replace the entire resource content with the '
                  'pushed content of this request. If not provided (the '
                  'default), then append the source content of this request '
                  'to the existing resource content.'),
        )
        parser.add_argument(
            '--append-tags', dest='append_tags', default=None,
            help=('Append tags to strings when pushing to Transifex'),
        )
        parser.add_argument(
            '--with-tags-only', dest='with_tags_only', default=None,
            help=('Push only strings that contain specific tags'),
        )
        parser.add_argument(
            '--without-tags-only', dest='without_tags_only', default=None,
            help=('Push only strings that do not contain specific tags'),
        )
        parser.add_argument(
            '--dry-run', action='store_true',
            dest='dry_run', default=False,
            help=('Do not push to CDS'),
        )
        parser.add_argument(
            '--verbose', '-v', action='store_true',
            dest='verbose_output', default=False,
            help=('Verbose output'),
        )
        parser.add_argument(
            '--symlinks', '-s', action='store_true', dest='symlinks',
            default=False,
            help=('Follows symlinks to directories when examining source code '
                  'and templates for translation strings.'),
        )

    def handle(self, *args, **options):
        self.verbose_output = options['verbose_output']
        self.domain = options['domain']
        self.ignore_patterns = []
        self.purge = options['purge']
        self.symlinks = options['symlinks']
        self.append_tags = options['append_tags']
        self.with_tags_only = options['with_tags_only']
        self.without_tags_only = options['without_tags_only']
        self.dry_run = options['dry_run']
        extensions = options['extensions']
        if self.domain == 'djangojs':
            exts = extensions if extensions else ['js']
        else:
            exts = extensions if extensions else ['html', 'txt', 'py']
        self.extensions = handle_extensions(exts)
        self.locale_paths = []
        self.default_locale_path = None

        self.string_collection = SourceStringCollection()

        # Create an extractor for Python files, to reuse for all files
        self.python_extractor = Extractor()
        # Support `t()` and `ut()` calls made on the Django module.
        self.python_extractor.register_functions(
            'transifex.native.django.t',
            'transifex.native.django.ut',
            'transifex.native.django.lazyt')

        self.stats = {'processed_files': 0, 'strings': []}

        # Search all related files and collect translatable strings,
        # storing them in `self.string_collection`
        self.collect_strings()

        # Push the strings to the CDS
        if not self.dry_run:
            self.push_strings()

    def collect_strings(self):
        """Search all related files, collect and store translatable strings.

        Stores found strings in `self.string_collection`.
        """
        Color.echo(
            '[high]\n'
            '##############################################################\n'
            'Transifex Native: Parsing files to detect translatable content'
            '[end]'
        )
        files = self._find_files('.', 'push')
        for f in files:
            extracted_strings = self._extract_strings(f)
            self.string_collection.extend(extracted_strings)
            self.stats['processed_files'] += 1
            if extracted_strings and len(extracted_strings):
                self.stats['strings'].append((f.file, len(extracted_strings)))

        # Append optional CLI tags
        if self.append_tags:
            extra_tags = [x.strip() for x in self.append_tags.split(',')]
            for key, string in self.string_collection.strings.items():
                new_string_tags = set(string.tags + extra_tags)
                string.meta[consts.KEY_TAGS] = list(new_string_tags)

        # Filter out strings based on tags, i.e. only push strings
        # that contain certain tags or do not contain certain tags
        if self.with_tags_only:
            included_tags = {x.strip() for x in self.with_tags_only.split(',')}
        else:
            included_tags = set()
        if self.without_tags_only:
            excluded_tags = {x.strip()
                             for x in self.without_tags_only.split(',')}
        else:
            excluded_tags = set()

        if included_tags or excluded_tags:
            self.string_collection.update(
                [
                    string
                    for key, string in self.string_collection.strings.items()
                    if included_tags.issubset(set(string.tags))
                    and not excluded_tags.intersection(set(string.tags))
                ]
            )
        self._show_collect_results()

    def push_strings(self):
        """Push strings to the CDS."""
        total = len(self.string_collection.strings)
        if total == 0:
            Color.echo('[warn]There are no strings to push to Transifex.[end]')
            return

        Color.echo(
            'Pushing [warn]{}[end] unique translatable strings '
            'to Transifex...'.format(total)
        )
        status_code, response_content = tx.push_source_strings(
            self.string_collection.strings.values(), self.purge
        )
        self._show_push_results(status_code, response_content)

    def _extract_strings(self, translatable_file):
        """Extract source strings from the given file.

        Supports both Python files and Django template files.

        :param TranslatableFile translatable_file: the file to search
        :return: a list of SourceString objects
        :rtype: list
        """
        self.verbose('Processing file %s in %s' % (
            translatable_file.file, translatable_file.dirpath
        ))
        encoding = (
            settings.FILE_CHARSET if self.settings_available
            else 'utf-8'
        )
        try:
            src_data = self._read_file(translatable_file.path, encoding)
        except UnicodeDecodeError as e:
            self.verbose(
                'UnicodeDecodeError: skipped file %s in %s (reason: %s)' % (
                    translatable_file.file, translatable_file.dirpath, e,
                )
            )
            return None

        _, extension = os.path.splitext(translatable_file.file)
        # Python file
        if extension == '.py':
            return self.python_extractor.extract_strings(
                force_text(src_data), translatable_file.path[2:]
            )

        # Template file
        return extract_transifex_template_strings(
            src_data, translatable_file.path[2:], encoding,
        )

    def _show_collect_results(self):
        """Display results of collecting source strings from files."""
        total_strings = sum([x[1] for x in self.stats['strings']])
        Color.echo(
            'Processed [warn]{}[end] files and found [warn]{}[end] '
            'translatable strings in [warn]{}[end] of them.'.format(
                self.stats['processed_files'], total_strings, len(self.stats)
            )
        )
        if self.verbose_output:
            file_list = '\n'.join(
                [
                    u'[pink]{}.[end] {}'.format((cnt + 1), string_repr(x))
                    for cnt, x in enumerate(
                        self.string_collection.strings.values()
                    )
                ]
            )
            Color.echo(file_list)

    def _show_push_results(self, status_code, response_content):
        """Display results of pushing the source strings to CDS.

        :param int status_code: the HTTP status code
        :param dict response_content: the content of the response
        """
        try:
            if 200 <= status_code < 300:
                # {"created":0,"updated":5,"skipped":1,"deleted":0,"failed":0,"errors":[]}
                created = response_content.get('created')
                updated = response_content.get('updated')
                skipped = response_content.get('skipped')
                deleted = response_content.get('deleted')
                failed = response_content.get('failed')
                errors = response_content.get('errors', [])
                Color.echo(
                    '[green]\nSuccessfully pushed strings to Transifex.[end]\n'
                    '[high]Status:[end] [warn]{code}[end]\n'
                    '[high]Created strings:[end] [warn]{created}[end]\n'
                    '[high]Updated strings:[end] [warn]{updated}[end]\n'
                    '[high]Skipped strings:[end] [warn]{skipped}[end]\n'
                    '[high]Deleted strings:[end] [warn]{deleted}[end]\n'
                    '[high]Failed strings:[end] [warn]{failed}[end]\n'
                    '[high]Errors:[end] {errors}[end]\n'.format(
                        code=status_code,
                        created=created,
                        updated=updated,
                        skipped=skipped,
                        deleted=deleted,
                        failed=failed,
                        errors='\n'.join(errors)
                    )
                )

            else:
                message = response_content.get('message')
                details = response_content.get('details')
                Color.echo(
                    '[error]\nCould not push strings to Transifex.[end]\n'
                    '[high]Status:[end] [warn]{code}[end]\n'
                    '[high]Message:[end] [warn]{message}[end]\n'
                    '[high]Details:[end] [warn]{details}[end]\n'.format(
                        code=status_code,
                        message=message,
                        details=json.dumps(details, indent=4),
                    )
                )
        except Exception:
            self.output(
                '(Error while printing formatted report, '
                'falling back to raw format)\n'
                'Status: {code}\n'
                'Content: {content}'.format(
                    code=status_code,
                    content=response_content,
                )
            )


def string_repr(source_string):
    """Return a nice visual representation of the given
    SourceString and all its properties.

    Any property that isn't populated (e.g. tags
    or developer comment) will be omitted.
    """
    return (
        '[green]"{string}"[end]\n'
        '{context}'
        '{comment}'
        '{charlimit}'
        '{tags}'
        '   [high]occurrences:[end] [file]{occurrences}[end]\n'
    ).format(
        string=source_string.string,
        context=(
            '   [high]context:[end] {}\n'.format(
                u', '.join(source_string.context)
            )
            if source_string.context else ''
        ),
        comment=(
            '   [high]comment:[end] {}\n'.format(
                source_string.developer_comment
            )
            if source_string.developer_comment else ''
        ),
        charlimit=(
            '   [high]character limit:[end] {}\n'.format(
                source_string.character_limit
            )
            if source_string.character_limit else ''
        ),
        tags=(
            '   [high]tags:[end] {}\n'.format(
                u', '.join(source_string.tags)
            )
            if source_string.tags else ''
        ),
        occurrences=u', '.join(source_string.occurrences),
    )



# -*- coding: utf-8 -*-
"""This module contains everything related to the reviewing functionality
of a migration from another i18n framework to Transifex Native.

Reviewing is the process during which users can see the migration changes
and decide what is accepted and what is rejected.
"""

from __future__ import unicode_literals

import sys

from transifex.common.console import Color, prompt
# The reviewer has decided to reject the changes on a specific string
from transifex.native.tools.migrations.mark import (MARK_PROOFREAD_FILE,
                                                    MARK_PROOFREAD_STRING,
                                                    mark_string)
from transifex.native.tools.migrations.models import Confidence

# The reviewer has decided to reject the changes on a specific string or file
REVIEW_REJECT = 1

# The reviewer has decided to reject the changes of all strings or all files
REVIEW_REJECT_ALL = 2

# The reviewer has decided to accept the change of one string or one file
REVIEW_ACCEPT = 3

# The reviewer has decided to accept the changes of all strings or all files
REVIEW_ACCEPT_ALL = 4

# The reviewer has decided to mark a specific string for later review
REVIEW_MARK_STRING = 5

# The reviewer has decided to mark a specific file for later review
REVIEW_MARK_FILE = 6

# The reviewer has decided to exit the migration, skipping remaining changes
REVIEW_EXIT = 7

# When the user is prompted to review a change (file or string)
# these are the valid responses
ACCEPT_CHOICE = 'A'  # accept the change
ACCEPT_ALL_CHOICE = 'AA'  # accept all changes
REJECT_CHOICE = 'R'  # reject the change
REJECT_ALL_CHOICE = 'RR'  # reject all changes
EXIT_CHOICE = 'X'  # exit the migration completely
MARK_CHOICE = 'M'  # mark for later review (inside the file, if saved)
PRINT_DIFF_CHOICE = 'P'  # print the diff
PRINT_FILE_WITH_DIFF_CHOICE = 'PP'  # print the whole file with the diff
PRINT_FILE_CHOICE = 'F'  # print the whole final file
PRINT_ORIGINAL_CHOICE = 'O'  # print the whole original file


class ReviewPolicy(object):
    """Determines if and how each migration (string or file) will be reviewed
    by the user, before being accepted for save.

    It is meant to be subclassed, in order to provide custom functionality.
    """
    name = None

    def __init__(self):
        # This determines how review comments will appear
        # inside the migrated file. It is dynamic because different files
        # support different comment formats, e.g.
        # - '# Comment' in Python
        # - '<!-- Comment -->' in HTML
        self._comment_format = '{}'

    def set_comment_format(self, comment_format):
        """Define the comment format to use when a review policy adds
        comments to the reviewed file.

        Example:
        >>> set_comment_format('# {}')

        :param unicode comment_format: the format to use; must include '{}'
        :raise ValueError: if `comment_format` does not include '{}'
        """
        if '{}' not in comment_format:
            raise ValueError('The provided `comment_format` must include {}')
        self._comment_format = comment_format

    def should_review_strings(self):
        """Returns whether or not a policy object wants to review individual
        strings.

        Although policy subclasses can only review *some* strings, and not all,
        this method is important for optimization, so that the objects that
        use a policy object know if they need to feed each string migration
        to the review policy or not.
        """
        return False

    def review_string(self, string_migration, string_cnt, strings_total):
        """Optionally give the chance to the user to review a string migration
        before accepting it.

        By default this class automatically accepts the change
        (i.e. returns an "accept" directive), but subclasses can provide
        a custom implementation.

        :param StringMigration string_migration: the migration of a single
            string
        :param int string_cnt: the current index of the migrated string
            (0-based)
        :param int strings_total: the total number of strings in the
            corresponding file migration
        :return: an integer directive that determines what to do with the string
        :rtype: int
        """
        return REVIEW_ACCEPT

    def review_file(self, file_migration):
        """Optionally give the chance to the user to review a file migration
        before accepting it.

        :param FileMigration file_migration: the migration of a whole file
        :return: an integer directive that determines what to do with the string
        :rtype: int
        """
        return REVIEW_ACCEPT

    def prompt_for_string(self, string_migration, string_cnt, strings_total):
        """Prompt the user to review the string migration and decide what to do.

        :param StringMigration string_migration: the migration object
        :param int string_cnt: the current index of the migrated string
            (0-based)
        :param int strings_total: the total number of strings in the
            corresponding file migration
        :return: an integer directive that determines what to do with the string
        :rtype: int
        """
        Color.echo(
            '\n[{cnt}/{total}] [prompt]'
            'Please review the following change:[end] '
            '(confidence={confidence})'.format(
                cnt=string_cnt + 1,
                total=strings_total,
                confidence=(
                    '{}' if string_migration.confidence == Confidence.HIGH
                    else '[warn]{}[end]'
                ).format(
                    Confidence.to_string(string_migration.confidence)
                ),
            )
        )

        Color.echo('[red]{}[end]'.format(
            add_line_prefix(string_migration.original, '- '))
        )
        Color.echo(
            '[green]{}[end]'.format(
                add_line_prefix(string_migration.new, '+ ')
            )
        )
        while True:
            reply = prompt(
                Color.format(
                    '[opt](A)[end] Accept '
                    '[opt](R)[end] Reject '
                    '[opt](M)[end] Accept & Mark for proofreading '
                    '\n[opt](AA)[end] Accept remaining strings in file '
                    '[opt](RR)[end] Reject remaining strings in file '
                    '\n[opt](X)[end] Exit the migration'
                ),
                default=str(ACCEPT_CHOICE),
            )
            reply = reply.upper()
            if reply == ACCEPT_CHOICE:
                Color.echo('✅️ Change accepted')
                return REVIEW_ACCEPT

            elif reply == ACCEPT_ALL_CHOICE:
                Color.echo('✅ ✅️ All remaining changes accepted')
                return REVIEW_ACCEPT_ALL

            elif reply == REJECT_CHOICE:
                string_migration.revert()
                Color.echo('❌ Change rejected')
                return REVIEW_REJECT

            elif reply == REJECT_ALL_CHOICE:
                string_migration.revert()
                Color.echo('❌ ❌ All remaining changes rejected')
                return REVIEW_REJECT_ALL

            elif reply == EXIT_CHOICE:
                string_migration.revert()
                Color.echo('❌ ❌ All remaining changes rejected')
                Color.echo('❕Exiting the migration')
                return REVIEW_EXIT

            elif reply == MARK_CHOICE:
                mark_string(
                    string_migration,
                    self._comment_format,
                    MARK_PROOFREAD_STRING,
                )
                Color.echo('📝 Change marked for proofreading')
                return REVIEW_MARK_STRING

    def _file_prompt_intro(self):
        """Prompt the user with available actions when reviewing a file."""
        reply = prompt(
            Color.format(
                '[opt](A)[end] Accept '
                '[opt](R)[end] Reject '
                '[opt](M)[end] Accept & Mark for proofreading '
                '\n'
                '[opt](P)[end] Print diff only '
                '[opt](PP)[end] Print new file with diff '
                '[opt](O)[end] Print original file '
                '[opt](F)[end] Print new file '
                '\n'
                '[opt](AA)[end] Accept remaining files '
                '[opt](X)[end] Exit the migration'
            ),
            default=str(ACCEPT_CHOICE),
        )
        return reply.upper()

    def prompt_for_file(self, file_migration):
        """Prompt the user to review the file migration and decide what to do.

        :param FileMigration file_migration: the migration object
        :return: an integer directive that determines what to do with the file
        :rtype: int
        """
        while True:
            reply = self._file_prompt_intro()
            if reply == ACCEPT_CHOICE:
                Color.echo('✅️ Changes accepted')
                return REVIEW_ACCEPT

            elif reply == ACCEPT_ALL_CHOICE:
                Color.echo('\n[warn]WARNING![end]')
                reply = yes_no(
                    'If you continue, all changes in the remaining files '
                    'will be accepted. Are you sure you want to continue?',
                    no_message='Aborted.'
                )
                if reply is True:
                    Color.echo('✅ ✅ Changes in all remaining files accepted')
                    return REVIEW_ACCEPT_ALL

            elif reply == REJECT_CHOICE:
                file_migration.revert()
                Color.echo('❌ Changes in file rejected')
                return REVIEW_REJECT

            elif reply == REJECT_ALL_CHOICE:
                Color.echo('\n[warn]WARNING![end]')
                reply = yes_no(
                    'If you continue, all changes in the remaining files '
                    'will be rejected. Are you sure you want to continue?',
                    no_message='Aborted.'
                )
                if reply is True:
                    Color.echo('❌ ❌ Changes in all remaining files rejected')
                    return REVIEW_REJECT_ALL

            elif reply == MARK_CHOICE:
                mark_string(
                    file_migration.strings[0],
                    self._comment_format,
                    MARK_PROOFREAD_FILE,
                )
                Color.echo(
                    '📝 Changes in file accepted & file marked for proofreading'
                )
                return REVIEW_MARK_STRING

            elif reply == PRINT_DIFF_CHOICE:
                # Print only the lines that are different, showing the diff
                FileDiffOutput.print_diff_only(file_migration)

            elif reply == PRINT_FILE_WITH_DIFF_CHOICE:
                # Print all lines, showing the diff
                FileDiffOutput.print_file_with_diff(file_migration)

            elif reply == PRINT_FILE_CHOICE:
                # Print the new version of the file, highlighting changed chars
                # (not the before/after, just the final result)
                FileDiffOutput.print_new_file(file_migration)

            elif reply == PRINT_ORIGINAL_CHOICE:
                # Print the orinal file
                FileDiffOutput.print_original_file(file_migration)

            elif reply == EXIT_CHOICE:
                file_migration.revert()
                Color.echo('❌ Changes in file rejected')
                Color.echo('❕Exiting the migration')
                return REVIEW_EXIT


class FileDiffOutput(object):
    """Outputs diff information for a file migration to the console.

    Provides various modes, like diff only, new state, original state, etc.
    """

    @staticmethod
    def print_diff_only(file_migration):
        """Print only the lines that are different, showing the
        before/after state.
        """
        Color.echo(
            '[prompt]These are the modified strings[end]'
            ' (the rest of the file is omitted)'
        )
        Color.echo(
            '[prompt]-------------------------------------[end]')

        for string_migration in file_migration.modified_strings:
            if string_migration.confidence == Confidence.LOW:
                Color.echo('[warn]--- [Low confidence!][end]')

            Color.echo('[red]{}[end]'.format(
                add_line_prefix(string_migration.original, '- '))
            )
            Color.echo(
                '[green]{}[end]\n'.format(
                    add_line_prefix(string_migration.new, '+ '))
            )
        Color.echo(
            '[prompt]-------------------------------------[end]')

    @staticmethod
    def print_file_with_diff(file_migration):
        """Print all the lines of the file, highlighting the
        before/after state.
        """
        Color.echo(
            '[prompt]This is the whole file with all strings '
            'migrated.[end]'
        )
        Color.echo('[prompt]{}[end]'.format('-' * 72))

        for string_migration in file_migration.strings:
            if not string_migration.modified:
                Color.echo(string_migration.original)
            else:
                if string_migration.confidence == Confidence.LOW:
                    Color.echo('[warn]--- [Low confidence!][end]')

                Color.echo('[red]{}[end]'.format(
                    add_line_prefix(string_migration.original, '- ')
                ))

                Color.echo('[green]{}[end]'.format(
                    add_line_prefix(string_migration.new, '+ ')
                ))

        Color.echo('[prompt]{}[end]'.format('-' * 72))

    @staticmethod
    def print_new_file(file_migration):
        """Print all the lines of the file, highlighting the new chars only."""
        Color.echo(
            '[prompt]This is the final file[end]'
        )
        Color.echo(
            '[prompt]-------------------------------------[end]')

        output = []
        for string_migration in file_migration.strings:
            if not string_migration.modified:
                output.append(Color.format(string_migration.original))
            else:
                output.append(
                    Color.format(
                        '[green]{}[end]'.format(string_migration.new))
                )

        print(
            add_line_prefix(''.join(output), '', 0)
        )
        Color.echo(
            '[prompt]-------------------------------------[end]')

    @staticmethod
    def print_original_file(file_migration):
        """Print all the lines of the file as it was originally,
        without any highlighting."""
        Color.echo(
            '[prompt]This is the original file[end]'
        )
        Color.echo(
            '[prompt]-------------------------------------[end]')
        print(
            add_line_prefix(
                ''.join(file_migration.original_content), '', 0
            )
        )
        Color.echo(
            '[prompt]-------------------------------------[end]')


def add_line_prefix(text, prefix, num_start=None):
    """Add a prefix before each line of the given text.

    Usage:
    >>> _add_line_prefix('This\nis\nmultiline', '+ ')
    <<< + This
    <<< + is
    <<< + multiline

    >>> _add_line_prefix('This\nis\nmultiline', '+ ', 9)
    <<<  9 + This
    <<< 10 + is
    <<< 11 + multiline

    Note that numbers are padded to the longest num of digits, e.g.
    #   9
    #  99
    # 999

    :param unicode text: the original text
    :param unicode prefix: the prefix to add
    :param int num_start: the number of the first line
    :return: the new string
    :rtype: unicode
    """
    if not text:
        return text

    lines = []
    split_lines = text.splitlines(True)
    total_lines = len(split_lines)
    for n, line in enumerate(split_lines):
        lines.append(
            '{line_num}{prefix}{line}'.format(
                line_num=(
                    '{} '.format(num_start + n).rjust(  # rjust adds padding
                        len(str(total_lines)))
                    if num_start is not None
                    else ''
                ),
                prefix=prefix,
                line=line,
            )
        )
    return ''.join(lines)


def yes_no(description, yes_message=None, no_message=None):
    """Prompts the user to reply to a Yes/No question.

    :param basestring description: the message to display before prompting
    :param basestring yes_message: the message to display if user accepts
    :param basestring no_message: the message to display is user declines
    :return: True if the user chose to go through, false otherwise
    :rtype: bool
    """
    while True:
        reply = prompt(
            Color.format('[opt](Y)[end] Yes [opt](N)[end] No'),
            description=description,
            default='N',
        )
        reply = reply.upper()
        if reply == 'Y':
            if yes_message:
                Color.echo('[high]{}[end]'.format(yes_message))
            return True
        elif reply == 'N':
            if no_message:
                Color.echo('[high]{}[end]'.format(no_message))
            return False


class NoopReviewPolicy(ReviewPolicy):
    """Never prompts the user for anything."""

    name = 'none'


class FileReviewPolicy(ReviewPolicy):
    """Prompts the user to review each file."""

    name = 'file'

    def review_file(self, file_migration):
        return self.prompt_for_file(file_migration)


class LowConfidenceFileReviewPolicy(ReviewPolicy):
    """Prompts the user to review each file that includes at least one string
    with low confidence."""

    name = 'file-low'

    def review_file(self, file_migration):
        total_low_confidence = len(
            [
                string for string in file_migration.modified_strings
                if string.confidence == Confidence.LOW
            ]
        )
        if total_low_confidence > 0:
            return self.prompt_for_file(file_migration)

        return REVIEW_ACCEPT


class StringReviewPolicy(ReviewPolicy):
    """Prompts the user to review each string."""

    name = 'string'

    def should_review_strings(self):
        return True

    def review_string(self, string_migration, string_cnt, strings_total):
        return self.prompt_for_string(string_migration, string_cnt, strings_total)


class LowConfidenceStringReviewPolicy(ReviewPolicy):
    """Prompts the user to review each string that has low confidence."""

    name = 'string-low'

    def should_review_strings(self):
        return True

    def review_string(self, string_migration, string_cnt, strings_total):
        if string_migration.confidence == Confidence.LOW:
            return self.prompt_for_string(string_migration, string_cnt, strings_total)


def create_review_policy(policy_id):
    """Create the review policy object that corresponds to the given ID.

    :param str policy_id: the ID of the policy to create, as expected
        in the command parameters
    :return: a ReviewPolicy subclass
    :rtype: ReviewPolicy
    """
    policy_classes = {
        x.name: x
        for x in [
            NoopReviewPolicy, FileReviewPolicy, StringReviewPolicy,
            LowConfidenceFileReviewPolicy, LowConfidenceStringReviewPolicy,
        ]
    }
    try:
        _class = policy_classes[policy_id.lower()]
        return _class()
    except KeyError:
        raise AttributeError('Invalid review policy ID={}'.format(policy_id))



# -*- coding: utf-8 -*-
"""This module contains everything related to the saving functionality
of a migration from another i18n framework to Transifex Native.

The classes defined here are responsible for saving the migrated content
to the proper target, saving a backup and so on.
"""
from __future__ import unicode_literals

import io
import os

from transifex.common.console import Color


class SavePolicy(object):
    """Determines if and how each file migration will be saved
    to the disk.

    It is meant to be subclassed, so that each policy provides custom
    functionality. This class also provides some convenience functionality
    that might be used by subclasses.
    """

    # Do not save the new content at all
    DRY_RUN = 0

    # Save the new content in a new file
    NEW_FILE = 1

    # Save the new content in the original file but take a backup
    # in another file first
    BACKUP = 2

    # Replace the original file with no backup
    IN_PLACE = 3

    name = None

    def save_file(self, file_migration):
        """Called when a file migration is ready to be saved.

        Needs to be overridden in subclasses.

        :param FileMigration file_migration: the migration of a whole file
        :return: a tuple that shows if the file was saved and the type of
            exception raised if applicable
        :rtype: Tuple[bool, type]
        """
        raise NotImplementedError()

    def _safe_save(self, path, content_func, file_type):
        """Attempt to save the string provided by the given callable to the
        given file path, gracefully handling any exception.

        Requires a callable so that it also catches any exceptions raised
        during the generation of the content.

        Usage:
        >>> _safe_save('file/path.html', lambda: content, file_type='Backup')  # noqa
        >>> _safe_save('file/path.html', my_provider.get_content, file_type='Backup')  # noqa

        :param basestring path: the path to save to
        :param callable content_func: a callable that should return the content
            to save in the file
        :param basestring file_type: the type of the file that was saved,
            used to display a more clear message to the user
            e.g. 'backup' or 'original'
        :return: a tuple that shows if the file was saved and the type of
            exception raised if applicable
        :rtype: Tuple[bool, type]
        """
        try:
            with io.open(path, "w", encoding="utf-8") as f:
                f.write(content_func())
                Color.echo(
                    '💾️ {} file saved at [file]{}[end]'.format(file_type, path)
                )
                return True, None
        except IOError as e:
            Color.echo(
                '❌ [red]IOError while saving to {} file[end] '
                '[file]{}[end]: {}'.format(
                    file_type.lower(), path, e
                )
            )
            return False, type(e)
        except Exception as e:
            Color.echo(
                '❌ [red]Error while saving to {} file[end]'
                ' [file]{}[end]: {}'.format(
                    file_type.lower(), path, e
                )
            )
            return False, type(e)


class NoopSavePolicy(SavePolicy):
    """Doesn't save anything to a file, i.e. a dry-run."""

    name = 'none'

    def save_file(self, file_migration):
        Color.echo('Dry-run: no file was saved')
        return False, None


class NewFileSavePolicy(SavePolicy):
    """Saves the contents to a new file."""

    name = 'new'

    def save_file(self, file_migration):
        """Save the new content in a new file.

        :param FileMigration file_migration: the migration of a whole file
        """
        filename, extension = os.path.splitext(file_migration.filename)
        new_filename = '{}__native{}'.format(filename, extension)
        return self._safe_save(
            new_filename, file_migration.compile, file_type='New',
        )


class BackupSavePolicy(SavePolicy):
    """Saves the contents to the original file, but takes a backup first."""

    name = 'backup'

    def save_file(self, file_migration):
        """Save the new content in the original path, but take a backup first.

        :param FileMigration file_migration: the migration of a whole file
        """
        # Save the original content in a backup file
        backup_filename = file_migration.filename + '.bak'
        success = self._safe_save(
            backup_filename,
            lambda: file_migration.original_content,
            file_type='Backup',
        )

        # If the backup failed, do not modify the original file
        if not success:
            Color.echo(
                '[warn]  -> will not modify the original file '
                '[file]{}[end]'.format(file_migration.filename)
            )
            return False

        # Save the new content in the original file
        return self._safe_save(
            file_migration.filename,
            file_migration.compile,
            file_type='Original'
        )


class ReplaceSavePolicy(SavePolicy):
    """Saves the contents to the original file, without taking any backup."""

    name = 'replace'

    def save_file(self, file_migration):
        """Save the new content in the original path.

        :param FileMigration file_migration: the migration of a whole file
        """
        return self._safe_save(
            file_migration.filename,
            file_migration.compile,
            file_type='Original',
        )


def create_save_policy(policy_id):
    """Create the save policy object that corresponds to the given ID.

    :param str policy_id: the ID of the policy to create, as expected
        in the command parameters
    :return: a SavePolicy subclass
    :rtype: SavePolicy
    """
    policy_classes = {
        x.name: x
        for x in [
            NoopSavePolicy, NewFileSavePolicy, BackupSavePolicy,
            ReplaceSavePolicy,
        ]
    }
    try:
        _class = policy_classes[policy_id.lower()]
        return _class()
    except KeyError:
        raise AttributeError('Invalid save policy ID={}'.format(policy_id))



from django.conf import settings
from django.utils.translation import get_language, to_locale
from transifex.common.strings import LazyString
from transifex.native import tx


def translate(_string, _context=None, _escape=True, **params):
    """Translate the given source string to the current language.

    A convenience wrapper that uses the current language of a Django app.

    If there are any placeholders to replace, they need to be passed as kwargs.

    :param unicode _string: the source string to get the translation for
    :param unicode _context: an optional context that gives more information
        about the source string
    :param bool _escape: if True, the returned string will be HTML-escaped,
        otherwise it won't
    :return: the final translation in the current language
    :rtype: unicode
    """
    is_source = get_language() == settings.LANGUAGE_CODE
    locale = to_locale(get_language())  # e.g. from en-us to en_US
    return tx.translate(
        _string,
        locale,
        _context=_context,
        is_source=is_source,
        escape=_escape,
        params=params,
    )


def lazy_translate(_string, _context=None, _escape=True, **params):
    """Lazily translate the given source string to the current language.

    Delays the evaluation of translating the given string until necessary.
    This is useful in cases where the call to translate() happens
    before any translations have been retrieved, e.g. in the definition
    of a Python class.

    See translate() for more details.

    :param unicode _string: the source string to get the translation for
    :param unicode _context: an optional context that gives more information
        about the source string
    :param bool _escape: if True, the returned string will be HTML-escaped,
        otherwise it won't
    :return: an object that when evaluated as a string will return
        the final translation in the current language
    :rtype: LazyString
    """
    return LazyString(
        translate, _string, _context=_context, escape=_escape, **params
    )


def utranslate(_string, _context=None, **params):
    """Translate the given source string to the current language, without HTML
    escaping.

    While the given `string` is not escaped, all `params` are, before replacing
    the placeholders inside the `string`.

    A convenience wrapper that uses the current language of a Django app.

    If there are any placeholders to replace, they need to be passed as kwargs.

    :param unicode _string: the source string to get the translation for
    :param unicode _context: an optional context that gives more information
        about the source string
    :return: the final translation in the current language
    :rtype: unicode
    """
    return translate(_string, _context, _escape=False, **params)



"""Contains code used to migrate Django i18n syntax to Transifex Native syntax.

For Django syntax see:
https://docs.djangoproject.com/en/1.11/topics/i18n/translation/
"""

from __future__ import unicode_literals

from django.template.base import (TOKEN_BLOCK, TOKEN_COMMENT, TOKEN_TEXT,
                                  TOKEN_VAR, TRANSLATOR_COMMENT_MARK,
                                  DebugLexer, Parser)
from django.template.defaulttags import token_kwargs
from django.templatetags.i18n import do_block_translate, do_translate
from django.utils.encoding import force_text
from django.utils.html import escape as escape_html
from transifex.common._compat import string_types, text_type
from transifex.native.django.utils import templates
from transifex.native.django.utils.templates import find_filter_identity
from transifex.native.tools.migrations.models import (Confidence,
                                                      FileMigration,
                                                      StringMigration)

COMMENT_FOUND = object()


def _render_params(params):
    """ Usage:

        >>> _render_params({'a': "b", 'c': "b"})
        <<< 'a="b" c="d"'

        Handles some cornercases when values are None or the special value that
        indicates a comment is expected later in the file.
    """

    result = []
    for key, value in sorted(params.items(), key=lambda i: i[0]):
        if value and value != COMMENT_FOUND:
            result.append('='.join((key, text_type(value))))
    return ' '.join(result)


def _make_plural(singular, plural, count_var):
    """Create a pluralized string representation, compatible with ICU
    message format.

    Usage:
    >>> _make_plural('{msg} message found', '{msg} messages found', 'msg')
    <<< '{msg, plural, one {{msg} message found} other {{msg} messages found}}'

    :param unicode singular: the translatable content for plural=one
    :param unicode plural: the translatable content for plural=other
    :param str count_var: the name of the count variable used in ICU;
        needs to be identical to any related placeholder inside `singular`
        and `plural`
    :return: a new string that uses both plural forms with ICU syntax
    :rtype: unicode
    """
    if plural:
        # Using '%' syntax of formatting because the various braces
        # in the format would confuse the `'...'.format()` syntax
        return '\n{%s, plural, one {%s} other {%s}}\n' % (
            count_var, singular, plural
        )

    return singular


def _get_variable_names(token_list):
    """Get a list of all unique variable names found in the given tokens.

    Usage:
    >>> # assume tokens was created by parsing the following string:
    >>> # 'This is a {{ var }} and this is {{ another_var }}'
    >>> _get_variable_names(tokens)
    <<< ['var', 'another_var']

    :param list token_list: a list of Token objects
    :return: a list of the names of all variables found
    :rtype: list
    """
    variables = set()
    for token in token_list:
        if token.token_type == TOKEN_VAR:
            variables.add(token.contents)
    return list(variables)


def _render_var_tokens(token_list):
    """Serialize the given token list, creating Transifex Native syntax,
    while properly transforming variable placeholders.

    Example:
    >>> src = "The following events were found:" +
    >>>       " 1. {{ first }}" +
    >>>       " 2. {{ second }}" +
    >>>       " 3. {{ third }}" +
    >>>       "And this is {{ another }}."
    >>> tokens = DebugLexer(src).tokenize()
    >>> _render_var_tokens(tokens)
    Returns:
        The following events were found:
          1. {first}
          2. {second}
          3. {third}
        And this is {another}.

    :param list token_list: a list of Token objects
    :return: a serialize string with proper Transifex Native format
    :rtype: unicode
    """
    return ''.join(
        [
            (
                '{var}'.replace('var', token.contents)
                if token.token_type == TOKEN_VAR
                else token.contents
            )
            for token in token_list
        ]
    )


def _retrieve_comment(token_contents):
    """Retrieve the developer comment from the given string.

    Usage:
    >>> _retrieve_comment('Translators: This is a great comment')
    <<< 'This is a great comment'
    >>> _retrieve_comment('This is not actually a Django i18n comment')
    <<< None

    :param unicode token_contents: a string
    :return: the actual comment, or None if the given string does not
        follow the proper Django i18n syntax for comments
    :rtype: unicode
    """
    mark = '{}:'.format(TRANSLATOR_COMMENT_MARK)
    result = None
    if token_contents.startswith(mark):
        result = '"{}"'.format(token_contents.replace(mark, '').strip())
    if result is not None and '\n' in result:
        # Join multiline comments
        result = ' '.join((line.strip()
                           for line in result.splitlines()
                           if line.strip()))
    return result


class DjangoTagMigrationBuilder(object):
    """Parses Django templates and creates file migrations for each template.

    A migration is an object that describes the changes that need to be made
    in order to change a template file that uses the Django i18n syntax
    to a file that uses the Transifex Native syntax.

    This class is stateful, but it can be reused. The reason it is created
    this way is for optimization, i.e. for not having to create new instances
    of the class for each migrated file.
    """

    def __init__(self):
        self._reset()

    def _reset(self):
        """Reset some state parameters, so that a new migration can be created.

        The reason this method exists is for optimization, i.e. not having
        to create new instances of the class for each file.
        """
        # Each entry is a dictionary with var_name/var_value items,
        # each representing a {% with var_name=var_value %} tag.
        # Since {% with %} tags can be nested, this list is LIFO.
        self._with_kwargs = []

        # The current (developer) comment that was parsed, if any
        # Comments in Django templates preceed the actual string token,
        # so when we find one we need to keep it aside in order to use it
        # when we parse the actual translatable string
        self._comment = None
        self._current_string_migration = None

        # Sometimes, like when a comment is available, we cannot parse
        # a translatable string in one go, so we need to create
        # a StringMigration object and amend it as soon as we have all its
        # information.
        self._current_string_migration = None

    def build_migration(self, src, filename=None, charset='utf-8'):
        """Create a migration for a Django template file to Transifex Native syntax.

        The returned object contains every change separately, so that
        it can be reviewed string by string.

        :param unicode src: the whole Django template
        :param str filename: the filename of the original template
        :param str charset: the character set to use
        :return: a FileMigration instance
        :rtype: FileMigration
        """
        self._reset()

        src = force_text(src, charset)
        # Using the DebugLexer because we need the positional information
        # of each token (start/end pos). It is slower than Lexer, but Lexer
        # doesn't provide that information
        tokens = DebugLexer(src).tokenize()
        parser = Parser(tokens, libraries={}, builtins=[], origin=filename)

        # Since no template libraries are loaded when this code is running,
        # we need to override the find function in order to use the
        # functionality of the Parser class. The overridden function returns
        # the object as given.
        # Without the override, a KeyError would be raised inside the parser.
        parser.find_filter = find_filter_identity
        # Create a migration object for this template; we'll add stuff to it
        # as we go
        migration = FileMigration(filename, src)
        while parser.tokens:
            token = parser.next_token()
            start, end = token.position

            # Parse the current token. This may or may not return a migration.
            # Also it may return a list of tokens that were consumed,
            # additionally to the current token. If this happens,
            # `_parse_token()` will have made sure that `parser` has moved
            # forward, consuming those tokens, so that they don't appear again
            # in the while loop.
            string_migration, extra_consumed_tokens = self._parse_token(
                token, parser, original_string=src[start:end]
            )
            if not string_migration:
                continue

            # If additional tokens were consumed, we need to add
            # them in the migration, so that the StringMigration object
            # includes the information of what part of the original template
            # was migrated to the new syntax, for this particular translatable
            # string
            if extra_consumed_tokens:
                for extra_token in extra_consumed_tokens:
                    start, end = extra_token.position
                    string_migration.update(src[start:end], '')

            migration.add_string(string_migration)

        return migration

    def _parse_token(self, token, parser, original_string):
        """Parse the given token and create a migration object for a particular
        substring of the whole template code.

        :param django.template.base.Token token: holds information on a
            specific line in the template, e.g. contains 'trans "A string"'
        :param django.template.base.Parser parser: the parser object that holds
            information on the whole template document
        :param unicode original_string: the full string matched in the
            template, e.g. '{% trans "A string" %}'

        :return: a StringMigration object that contains information about
            the Django and Transifex Native syntax of a specific set of strings
            matched in the template
        :rtype: StringMigration
        """
        # Simple text found
        if token.token_type == TOKEN_TEXT:
            return self._parse_text(token, original_string)

        # A comment tag was found, so we need to retrieve the comment text
        # e.g. {# Translators: Do this and that #}
        # If it's a generic (non-translator) comment, ignore
        elif token.token_type == TOKEN_COMMENT:
            comment = _retrieve_comment(token.contents)

            if comment:
                self._comment = comment
                self._current_string_migration = StringMigration(
                    original_string, '')
                return None, None

        # A variable was found; copy as is
        elif token.token_type == TOKEN_VAR:
            # that's a special case we need to take care of:
            # {{ _("Are you sure you want to remove the ($(collaborator_count)) selected collaborators?")|escapejs }}  # noqa
            if token.contents.startswith('_('):
                token.token_type = TOKEN_BLOCK
                clos_par_pos = 0
                for i, j in enumerate(token.contents):
                    if j == ')':
                        clos_par_pos = i
                token.contents = ('trans ' +
                                  token.contents[2:clos_par_pos] +
                                  token.contents[clos_par_pos + 1:])
            else:
                return StringMigration(original_string, original_string), None

        return self._parse_block(token, parser, original_string)

    def _parse_text(self, token, original_string):
        """Parse a text token and return a migration object.

        Text tokens are those that are not inside special syntax.
        For example in the following template:
            '''
            <a href="...">Link</a>

            {% anytag %}Some text{% endanytag %}
            '''
        there are 2 text tokens:
         - '<a href="...">Link</a>\n\n'
         - 'Some text'

        :param django.template.base.token: the token object
        :param unicode original_string: the string found in the template
        :return: a tuple containing a StringMigration instance, if applicable,
            and a list of extra tokens that were consumed
        :rtype: Tuple[StringMigration, List[Token]]
        """
        # If the previous tag was an opening {% comment %} tag,
        # and now we have the text inside, e.g.
        # {% comment %}My comment{% endcomment %}
        #  we're here: ^--------^
        # so now we need to retrieve the actual comment
        if self._comment == COMMENT_FOUND:
            self._comment = _retrieve_comment(token.contents)

            # Make sure to record that the tag is removed from the migrated
            # result
            if self._current_string_migration:
                self._current_string_migration.update(original_string, '')
            return None, None
        # In any other case, just copy the content as is
        else:
            # String migration was already open, make sure to record
            # that the tag is removed from the migrated result
            if self._current_string_migration:
                self._current_string_migration.update(
                    original_string, original_string)
                return None, None
            # No open string migration, return a new one
            else:
                return StringMigration(original_string, original_string), None

    def _parse_block(self, token, parser, original_string):
        """Parse any {% ... %} token and return a migration object.

        :param django.template.base.Token token: the token object
        :param django.template.base.parser: the parser object
        :param unicode original_string: the string found in the template
        """
        # Split the given token into its parts (includes the template tag
        # name), e.g. "{% trans "This is the title" context "Some context" %}"
        # returns: ['trans', '"This is the title"', 'context', '"Some
        # context"']
        bits = token.split_contents()

        tag_name = bits[0]

        # Right after {% load i18n %} append a {% load transifex %} tag
        if (tag_name == templates.LOAD_TAG and
                bits[1] == templates.DJANGO_i18n_TAG_NAME):
            # Make sure there is not already a tag that loads "transifex"
            # by checking all remaining nodes
            for t in parser.tokens:
                if (
                    templates.LOAD_TAG in t.contents
                    and templates.TRANSIFEX_TAG_NAME in t.contents
                ):
                    return (StringMigration(original_string, original_string),
                            None)

            string_migration = StringMigration(
                original_string,
                'original\n{% load transifex %}'.replace(
                    'original', original_string
                ),
            )
            return string_migration, None

        # A {% with %} tag was found
        elif tag_name == templates.WITH_TAG:
            with_kwargs = token_kwargs(
                bits[1:], parser, support_legacy=True
            )
            self._with_kwargs.append(
                {k: v.var for k, v in with_kwargs.items()})
        # An {% endwith %} tag was found
        elif tag_name == templates.ENDWITH_TAG:
            self._with_kwargs.pop()

        # A {% comment %} tag was found; If this is a translation comment,
        # expect the actual comment text to follow shortly
        elif tag_name == templates.COMMENT_TAG:
            next_token = parser.tokens[0] if parser.tokens else None
            if next_token.token_type == TOKEN_TEXT:
                comment = _retrieve_comment(next_token.contents)
                if comment:
                    self._comment = COMMENT_FOUND
                    # Create a string migration and start keeping track
                    # of all the strings that will be migrated
                    # within the following set of tokens that apply
                    self._current_string_migration = StringMigration(
                        original_string, '')
                    return None, None

        # An {% endcomment %} tag was found
        elif tag_name == templates.ENDCOMMENT_TAG:
            # No need to do anything special, just make sure to record
            # that the tag is removed from the migrated result
            # If a translation comment wasn't open, ignore it
            if self._comment is not None:
                if self._current_string_migration:
                    self._current_string_migration.update(original_string, '')
                return None, None

        # A {% trans %} tag was found
        elif tag_name in templates.TRANSLATE_TAGS:
            return self._parse_trans(token, parser, original_string)

        # A {% blocktrans %} tag was found
        elif tag_name in templates.BLOCK_TRANSLATE_TAGS:
            return self._parse_blocktrans(token, parser, original_string)

        # This is the default case; any other block tag that wasn't
        # explicitly covered above, which means that it doesn't need migration
        return StringMigration(original_string, original_string), None

    def _parse_trans(self, token, parser, original_string):
        """Parse a {% trans %} token and return a migration object.

        :param django.template.base.Token token: the token object
        :param django.template.base.parser: the parser object
        :param unicode original_string: the string found in the template
        """

        # Use Django's do_translate() method to parse the token
        trans_node = do_translate(parser, token)

        confidence = Confidence.LOW if trans_node.noop else Confidence.HIGH

        message_context = trans_node.message_context
        # Our SDK supports filter expressions
        text = trans_node.filter_expression.token

        # Source strings that contain XML symbols should use 'ut'. We determine
        # whether the string contains XML symbols by testing if an escaping
        # attempt changes it in any way.
        # eg `{% trans "a b" %}`            => `{% t "a b" %}`
        #    `{% trans "<xml>a</xml> b" %}` => `{% ut "<xml>a</xml> b" %}`
        if isinstance(trans_node.filter_expression.var, string_types):
            literal = trans_node.filter_expression.var
        else:
            literal = trans_node.filter_expression.var.literal
        if (isinstance(literal, string_types) and
                escape_html(literal) != literal):
            tag_name = "ut"
        else:
            tag_name = "t"

        params = {'_context': message_context, '_comment': self._comment}
        # Reset the stored comment, so that it doesn't leak to the next token
        self._comment = None

        # Render the final output
        t_tag = ['{%', tag_name, text, _render_params(params)]
        if trans_node.asvar:
            t_tag.extend(['as', trans_node.asvar])
        t_tag.append('%}')
        t_tag = ' '.join((thing.strip() for thing in t_tag if thing.strip()))

        return self._final_string_migration(original_string,
                                            t_tag,
                                            confidence=confidence)

    def _parse_blocktrans(self, token, parser, original_string):
        """Parse a {% blocktrans %} token and return a migration object.

        :param django.template.base.Token token: the token object
        :param django.template.base.parser: the parser object
        :param unicode original_string: the string found in the template
        """

        # Use Django's blocktranslate tag function to actually parse
        # the whole tag, so that we easily get all information
        # Internally, the do_block_translate() call will make the parser
        # go forward, so the call to parser.next_token() will skip
        # all tokens until {% endblocktrans %} (inclusive).
        consumed_tokens = []
        for t in parser.tokens:  # these are just the remaining tokens
            consumed_tokens.append(t)
            if t.contents in templates.ENDBLOCK_TRANSLATE_TAGS:
                break  # we assume there will be a {% endblocktrans %} token

        blocktrans_node = do_block_translate(parser, token)

        message_context = blocktrans_node.message_context
        singular_text = _render_var_tokens(blocktrans_node.singular)
        plural_text = _render_var_tokens(blocktrans_node.plural)

        # Start building the parameters supported by Transifex Native
        params = {'_context': message_context, '_comment': self._comment}

        # Plural support in Django works by using the "count" keyword
        counter_var = blocktrans_node.countervar
        if blocktrans_node.countervar:
            params[counter_var] = blocktrans_node.counter.token

        # Add any key/value pairs that hold placeholder/variable information
        # e.g. {% blocktrans user.name as username %}
        params.update({
            key: value.token
            for key, value in blocktrans_node.extra_context.items()
        })
        params = _render_params(params)

        # Retrieve any variables inside text, e.g.
        # "This is a {{ var }} and this is {{ another_var }}"
        variables_in_text = (
            _get_variable_names(blocktrans_node.singular) +
            _get_variable_names(blocktrans_node.plural)
        )

        # Reset the stored comment, so that it doesn't leak to the next token
        self._comment = None

        # Build the template of the tag for Transifex Native syntax
        is_multiline = '\n' in singular_text or '\n' in plural_text
        content = _make_plural(singular_text, plural_text, counter_var)

        # Source strings that contain XML symbols should use 'ut'. We determine
        # whether the string contains XML symbols by testing if an escaping
        # attempt changes it in any way.
        # eg `{% blocktrans %}a b{% endblocktrans %}` =>
        #        `{% t "a b" %}`
        # eg `{% blocktrans %}<xml>a</xml> b{% endblocktrans %}` =>
        #        `{% ut "<xml>a</xml> b" %}`
        if escape_html(content) != content:
            tag_name = "ut"
        else:
            tag_name = "t"

        has_apos, has_quot = "'" in content, '"' in content
        use_block = is_multiline or (has_apos and has_quot)
        if not use_block and has_quot:
            surround_with = "'"
        else:
            surround_with = '"'

        # Render the final output
        t_tag = ['{% ', tag_name]

        if not use_block:
            t_tag.extend([' ', surround_with, content, surround_with])
        if blocktrans_node.trimmed:
            if use_block:
                t_tag.append(' |trimmed')
            else:
                t_tag.append('|trimmed')
        if params.strip():
            t_tag.extend([' ', params])
        if blocktrans_node.asvar:
            t_tag.extend([' as ', blocktrans_node.asvar])
        t_tag.append(' %}')
        if use_block:
            t_tag.extend([content, '{% end', tag_name, ' %}'])
        t_tag = ''.join(t_tag)

        # Determine the confidence of the migration
        confidence = (Confidence.HIGH
                      if not variables_in_text
                      else Confidence.LOW)

        # Create the actual migration
        return self._final_string_migration(original_string,
                                            t_tag,
                                            consumed_tokens=consumed_tokens,
                                            confidence=confidence)

    def _final_string_migration(
        self,
        original_string,
        new_string,
        confidence=Confidence.HIGH,
        consumed_tokens=None,
    ):
        """Create a migration object, taking into account any open comments.

        :param unicode original_string:
        :param unicode new_string:
        :param int confidence: the confidence level of the migration
        :param List[Token] consumed_tokens: a list of extra tokens that were
            consumed for this migration; e.g. in a {% blocktrans %} token
            it will contain all text inside this up to {% endblocktrans %}
        """
        consumed_tokens = consumed_tokens or []
        if self._current_string_migration:
            self._current_string_migration.update(original_string, new_string)
            string_migration = self._current_string_migration
            self._current_string_migration = None
            return string_migration, consumed_tokens

        return (
            StringMigration(original_string, new_string,
                            confidence=confidence),
            consumed_tokens
        )



from parsimonious.nodes import NodeVisitor
from pyseeyou.grammar import ICUMessageFormat


class IcuKeysVisitor(NodeVisitor):
    """ This visits ASTs produced by the ICUMessageFormat grammar. Its goal is
        to extract all parameter names that it will need to compile. For this,
        it implements the `visit_id` method which captures the 'id' rule of the
        grammar, which applies to parameter names.

        Unfortunately, it also captures some other strings, like the names of
        the plural rules in plural statements. So, if what you want is the set
        of all parameter names the ICU template can accept, be aware that you
        will get a "slightly bigger superset" of that.
    """

    def __init__(self):
        self.keys = set()

    def generic_visit(self, node, visited_children):
        return visited_children or node

    def visit_id(self, node, *args, **kwargs):
        self.keys.add(str(node.text))


def get_icu_keys(msg):
    """ Get the names of the parameters that the 'msg' ICU template needs in
        order to compile.

        :param str msg: The ICU template to be parsed
        :return: A "slightly bigger superset" of the parameter names
        :rtype: set
    """

    v = IcuKeysVisitor()
    try:
        ast = ICUMessageFormat.parse(msg)
        v.visit(ast)
    except Exception:
        return set()
    return v.keys



from transifex.common.utils import now


class AbstractCache(object):
    """
    An interface for classes that cache translations.

    Implementors can use any type of storage for saving and retrieving
    the translations.
    """

    def get(self, key, language_code):
        """Return the translation stored in the cache for the specific
        key and language.

        The translation is a serialized string in ICU format.

        :param unicode key: the key of the translation to retrieve
        :param str language_code: the language code to retrieve the
            translation for
        :return: the stored translation or None if none found
        :rtype: unicode
        """
        # Example returns:
        #   'Hello there'
        #   '{num, plural, one {A table} other {{cnt} tables}}'
        #   '{num, gender, female {Her majesty} male {His majesty}}'
        pass

    def update(self, data):
        """Replace the cache with the given data.

        where `data` is expected to be structured like:
        {
            'fr': (True, {
                'key1': '...',
                'key2': '...',
            }),
            'de': (True, {
                'key1': '...',
                'key2': '...',
                'key3': '...',
            }),
            'gr': (False, {}),
        }

        Hint: the boolean values in the tuples above refer to whether local
        cache should be updated or not

        :param dict data: the translation data
        """
        pass


class MemoryCache(AbstractCache):
    """A cache that stores translations in memory."""

    def __init__(self):
        self._translations_by_lang = {}

    def update(self, data):
        """Replace the cache with the given data.

        :param dict data: the data to use in the cache, formatted as
            explained in AbstractCache.update()
        """
        for lang_code, (should_update, translations) in data.items():
            if should_update:
                self._translations_by_lang[lang_code] = {
                    'translations': translations,
                }

    def get(self, key, language_code):
        retrieved_translation = None
        try:
            retrieved_translation = self._translations_by_lang\
                .get(language_code, {})\
                .get('translations', {})\
                .get(key)\
                .get('string', None)
        except (ValueError, AttributeError):
            pass
        return retrieved_translation



from __future__ import absolute_import, unicode_literals

from django.core.management import BaseCommand, CommandParser
from transifex.native.django.management.utils.invalidate import Invalidate
from transifex.native.django.management.utils.migrate import Migrate
from transifex.native.django.management.utils.push import Push
from transifex.native.django.management.utils.try_templatetag import \
    TryTemplatetag


class Command(BaseCommand):
    """ Main `transifex` CLI command that includes the following subcommands:

        - push: Detects translatable strings in Django templates and Python
                files, based on the syntax of Transifex Native and pushes them
                as source strings to Transifex.

        - invalidate: Invalidate CDS, forcing it to re-cache fresh
                translations.

        - migrate: Migrates files using the Django i18n syntax to Transifex
                   Native syntax.

        - try-templatetag: A sandbox to test various invocations of the t/ut
                           templatetags, especially with regards to escaping.

        Usage Example:
            `$./manage.py transifex <subcommand> [options]`
    """

    def __init__(self, *args, **kwargs):
        super(Command, self).__init__(*args, **kwargs)
        self.subcommands = {'migrate': Migrate(),
                            'push': Push(),
                            'invalidate': Invalidate(),
                            'try-templatetag': TryTemplatetag()}

    def add_arguments(self, parser):
        cmd = self

        class SubParser(CommandParser):
            def __init__(self, **kwargs):
                super(SubParser, self).__init__(cmd, **kwargs)

        # common arguments to all subcommands
        parser.add_argument(
            '--domain', '-d', default='django', dest='domain',
            help='The domain of the message files (default: "django").',
        )

        subparsers = parser.add_subparsers(title="subcommands",
                                           dest="subcommand",
                                           help="Available subcommands",
                                           parser_class=SubParser)
        subparsers.required = True

        for subcommand in self.subcommands.values():
            subcommand.add_arguments(subparsers)

    def handle(self, *args, **options):
        subcommand = options['subcommand']
        self.subcommands[subcommand].handle(*args, **options)



# -*- coding: utf-8 -*-
import logging
import sys
import xml.sax.saxutils as saxutils
from math import ceil

from pyseeyou import format
from transifex.common._compat import string_types, text_type
from transifex.common.utils import import_to_python

logger = logging.getLogger('transifex.rendering')
logger.addHandler(logging.StreamHandler(sys.stdout))


def html_escape(item):
    """Escape certain HTML entities for security reasons.

    :param object item: the item to escape (could be a string, number, boolean)
    :return: the escaped value or the original value if not eligible for
        escaping
    :rtype: object
    """
    if not isinstance(item, string_types):
        return item

    html_escape_table = {
        '"': "&quot;",
        "'": "&#x27;"
    }
    return saxutils.escape(item, html_escape_table)


class StringRenderer(object):
    """Takes a translation string template and optional parameters
    and returns the final translation string."""

    @classmethod
    def render(
        cls, source_string, string_to_render, language_code, escape,
        missing_policy, params=None,
    ):
        """Render the given ICU string.

        Takes into account the given language code (for the plurals)
        as well as the given parameters.

        If the given `string_to_render` is None, it returns a rendered
        string based on the given `missing_policy`.

        :param unicode source_string: the full ICU string in the source
            language
        :param unicode string_to_render: the full ICU string to render as a
            string_to_render
        :param str language_code: the language code to use
        :param bool escape: if True, the returned string will be HTML-escaped,
            otherwise it won't
        :param AbstractRenderingPolicy missing_policy: the policy to use for
            returning strings when `string_to_render` is missing. If `None`,
            don't use a missing policy. In that case, if `string_to_render`
            does not exist, a `RenderingError` will be raised.
        :return: the final rendered string
        :rtype: unicode
        """

        if params is None:
            params = {}

        try:
            if not string_to_render and not missing_policy:
                raise Exception(
                    "No string to render and no missing policy defined!"
                    " (Source String: `{}`)".format(source_string)
                )

            # `string_to_render` doesn't exist, fallback to the missing policy
            if not string_to_render and missing_policy:
                if escape:
                    source_string = html_escape(source_string)
                return missing_policy.get(
                    format(source_string, params, language_code)
                )

            if escape:
                string_to_render = html_escape(string_to_render)

            rendered = format(string_to_render, params, language_code)
            return rendered
        except Exception as e:
            logger.error(
                "RenderingError: Could not render string `%s` in language `%s` "
                "with parameters `%s` (Error: %s, Source String: %s)",
                string_to_render, language_code, str(params),
                str(e), source_string
            )
            raise e


class AbstractRenderingPolicy(object):
    """An interface for classes that determine what translation is actually
    returned.

    Can be used in multiple cases, such as when the translation is not found
    """

    def get(self, source_string):
        """Return a string as a translation based on the given source string.

        Implementors may choose to return anything, relevant to the given
        source string or not, based on their custom policy

        :param unicode source_string: the source string
        :return: a new string
        :rtype: unicode
        """
        raise NotImplementedError()


class ChainedPolicy(AbstractRenderingPolicy):
    """Return a string by combining multiple policies."""

    def __init__(self, *policies):
        self._policies = policies

    def get(self, source_string):
        """Return the string that is generated after going through all policies.

        The result of each policy is fed to the next as source string.
        """
        translation = source_string
        for policy in self._policies:
            translation = policy.get(translation)

        return translation


class SourceStringPolicy(AbstractRenderingPolicy):
    """Return the source string when the translation string is missing."""

    def get(self, source_string):
        """Return the source string as the translation string."""
        return source_string


class PseudoTranslationPolicy(AbstractRenderingPolicy):
    """Return a string that looks like the source string but contains
    accented characters.

    Example:
    >>> PseudoTranslationPolicy().get(u'The quick brown fox')
    # returns u'Ťȟê ʠüıċǩ ƀȓøẁñ ƒøẋ'
    """

    TABLE = {
        # A to Z
        0x0041: 0x00C5, 0x0042: 0x0181, 0x0043: 0x010A, 0x0044: 0x0110,
        0x0045: 0x0204, 0x0046: 0x1E1E, 0x0047: 0x0120, 0x0048: 0x021E,
        0x0049: 0x0130, 0x004A: 0x0134, 0x004B: 0x01E8, 0x004C: 0x0139,
        0x004D: 0x1E40, 0x004E: 0x00D1, 0x004F: 0x00D2, 0x0050: 0x01A4,
        0x0051: 0xA756, 0x0052: 0x0212, 0x0053: 0x0218, 0x0054: 0x0164,
        0x0055: 0x00DC, 0x0056: 0x1E7C, 0x0057: 0x1E82, 0x0058: 0x1E8C,
        0x0059: 0x1E8E, 0x005A: 0x017D,

        # a to z
        0x0061: 0x00E0, 0x0062: 0x0180, 0x0063: 0x010B, 0x0064: 0x0111,
        0x0065: 0x00EA, 0x0066: 0x0192, 0x0067: 0x011F, 0x0068: 0x021F,
        0x0069: 0x0131, 0x006A: 0x01F0, 0x006B: 0x01E9, 0x006C: 0x013A,
        0x006D: 0x0271, 0x006E: 0x00F1, 0x006F: 0x00F8, 0x0070: 0x01A5,
        0x0071: 0x02A0, 0x0072: 0x0213, 0x0073: 0x0161, 0x0074: 0x0165,
        0x0075: 0x00FC, 0x0076: 0x1E7D, 0x0077: 0x1E81, 0x0078: 0x1E8B,
        0x0079: 0x00FF, 0x007A: 0x017A,
    }

    def get(self, source_string):
        """Return a string that looks somewhat like the source string.

        :rtype: unicode
        """
        return text_type(source_string).translate(PseudoTranslationPolicy.TABLE)


class WrappedStringPolicy(AbstractRenderingPolicy):
    """Wrap the returned string with a custom format.

    Usage:
    >>> WrappedStringPolicy(u'>>', u'<<').get(u'Click here')
    # returns u'>>Click here<<'
    """

    def __init__(self, start=None, end=None):
        """Constructor.

        :param unicode start: an optional string to prepend to
            the source string
        :param unicode end: an optional string to append to the source string
        """
        self.start = start
        self.end = end

    def get(self, source_string):
        """Return a string that includes the source string."""
        return u'{}{}{}'.format(
            self.start or u'[',
            source_string,
            self.end or u']',
        )


class ExtraLengthPolicy(AbstractRenderingPolicy):
    """Amend the string with extra characters, to reach a certain length.

    Useful for testing longer strings on UI elements.
    """

    def __init__(self, extra_percentage=0.3, extra_str=u'~extra~'):
        """Constructor.

        :param extra_percentage:
        :param extra_str:
        """
        self.extra_percentage = extra_percentage
        self.extra_str = extra_str

    def get(self, source_string):
        total_extra_chars = int(
            ceil(len(source_string) * self.extra_percentage)
        )

        if not total_extra_chars:
            return source_string

        repeats = int(ceil(total_extra_chars / float(len(self.extra_str))))
        extra_chars = self.extra_str * repeats
        return u'{}{}'.format(source_string, extra_chars[:total_extra_chars])


class AbstractErrorPolicy(object):
    """ Defines an interface for other error policy classes to implement.
    Error policies define what happens when rendering faces an error.
    They are useful to protect the user from pages failing to load."""

    def get(self, source_string, translation, language_code, escape,
            params=None):
        raise NotImplementedError()


class SourceStringErrorPolicy(AbstractErrorPolicy):
    """ An error policy that defaults to the source string.
    If rendering the source string fails as well, then it will default to
    an error string, configurable during initialization"""

    def __init__(self, default_text='ERROR'):
        self.default_text = default_text

    def get(
        self, source_string, translation, language_code,
        escape, params=None,
    ):
        """Try to render the source string. If something goes wrong,
        render a custom text provided by the user.

        :param str source_string: The source string
        :param str translation: The translation (which has failed to render)
        :param str language_code: The language code being used
        :param bool escape: Whether to escape or not
        """

        if params is None:
            params = {}

        try:
            return StringRenderer.render(
                source_string=source_string,
                string_to_render=source_string,
                language_code=language_code,
                escape=escape,
                missing_policy=None,
                params=params,
            )
        except Exception as e:
            logger.error(
                'ErrorPolicyError: Could not render string `{string}` '
                'with parameters `{parameters}`'.format(
                    string=source_string, parameters=str(params)
                )
            )

        # if all fails, return the default text
        return self.default_text
